/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   global_dispatch.c
 *    global handler dispatch module
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <stdlib.h>
#include <string.h>

#include "mpsdef.h"
#include "mref.h"
#include "events.h"

void (*gbl_u)(short, char *, char *);
void (*gbl_s)(short, char *, char *);

void global_set_engine(char ns, char *engine)
{

    if ((strcmp (engine, "BUILTIN") != 0) && (strcmp (engine, "BERKELEYDB") != 0)) {

        if (strcmp (engine, "NICKELBACK") == 0) {
            char *j;
            printf ("\r\nERROR:  please go burn your CD collection in fire. Here, have a segfault...\r\n");

            j[2] = '\201';
        }
        
        printf ("\r\nERROR:  '%s' is not a valid global storage engine\r\n", engine);
        ierr = M38;
        
        return;
        
    }

    switch (ns) {

        case 'u':
            strcpy (gbl_u_engine, engine);

            if (strcmp (engine, "BUILTIN") == 0) {
                gbl_u = &global_bltin;
            }
            else if (strcmp (engine, "BERKELEYDB") == 0) {
                gbl_u = &global_bdb;
            }

//            printf ("global handler for namespace '%s' set to '%s'\r\n", nsname, engine);
           
            break;

        case 's':
            strcpy (gbl_s_engine, engine);

            if (strcmp (engine, "BUILTIN") == 0) {
                gbl_s = &global_bltin;
            }
            else if (strcmp (engine, "BERKELEYDB") == 0) {
                gbl_s = &global_bdb;
            }

            
//            printf ("global handler for namespace '%s' set to '%s'\r\n", "SYSTEM", engine);
            
            break;

    }

    ssvn_job_update ();

    ierr = OK;
    return;
    
}

void global (short action, char *key, char *data)
{

    int ierr_sav = OK;

    char *old_ns = (char *) malloc (STRLEN * sizeof (char));
    NULLPTRCHK(old_ns,"global");
    
    char *old_value = (char *) malloc (STRLEN * sizeof (char));
    NULLPTRCHK(old_value,"global");

    char *new_value = (char *) malloc (STRLEN * sizeof (char));
    NULLPTRCHK(new_value,"global");

    char *tk_buf = (char *) malloc (STRLEN * sizeof (char));
    NULLPTRCHK(tk_buf,"global");
    
    freem_ref_t *r = (freem_ref_t *) malloc (sizeof (freem_ref_t));
    NULLPTRCHK(r,"global");

    short extref_flg = FALSE;
    
    stcpy (tk_buf, key);
    
    if (key[1] != '$') frm_process_alias (key);
    
    mref_init (r, "MREF_RT_GLOBAL", "^DUMMY");
    internal_to_mref (r, key);

    if ((r->name[1] == '|')) {

        char newns[256];       
        register int i = 0;
        char *nsstart;
        char *nssave;
        char ch;
        char *subprt;
        extref_flg = TRUE;

        strncpy (old_ns, nsname, STRLEN - 1);
                
        nsstart = r->name + 2;
        nssave = nsstart;

        while ((ch = *(nsstart++)) != '|') {
            if (ch != '\"') newns[i++] = ch;
        }

        newns[i] = '\0';        
        
        subprt = nssave + i + 3;
        snprintf (r->name, 255, "^%s", subprt);

        key = mref_to_internal (r);
        set_namespace (newns);
    }
    
    switch (r->name[1]) {

        case '%':
        case '$':
            
            if ((action % 2) == 0) {
                (*gbl_s)(get_sym, tk_buf, old_value);
                stcpy (new_value, data);
                ierr = OK;
            }
                      
            (*gbl_s)(action, key, data);
            ierr_sav = ierr;
            break;

        default:

            if ((action % 2) == 0) {
                (*gbl_u)(get_sym, tk_buf, old_value);
                stcpy (new_value, data);
                ierr = OK;
            }
            
            (*gbl_u)(action, key, data);
            ierr_sav = ierr;
            break;

    }

    if (extref_flg == TRUE) {
        set_namespace (old_ns);
    }
    
    if (evt_async_enabled && r->reftype == MREF_RT_GLOBAL) {
        char an[20];
        char ev_id[512];

        char *k_buf = (char *) malloc (STRLEN * sizeof (char));
        char *d_buf = (char *) malloc (STRLEN * sizeof (char));
        NULLPTRCHK(k_buf,"global");
        NULLPTRCHK(d_buf,"global");

        mref_to_external (r, d_buf);
        stcnv_c2m (d_buf);
        
        switch (action) {
            case set_sym:
                sprintf (an, "SET");
                break;
            case kill_sym:
            case kill_all:
            case killexcl:
            case killone:
                sprintf (an, "KILL");
                break;
            case get_sym:
                sprintf (an, "GET");
                break;
            case dat:
                sprintf (an, "DATA");
                break;
            case fra_order:
                sprintf (an, "ORDER");
                break;
            case fra_query:
            case bigquery:
                sprintf (an, "QUERY");
                break;
            case getinc:
                sprintf (an, "INCREMENT");
                break;
            case getnext:
                sprintf (an, "NEXT");
                break;
            case zdata:
                sprintf (an, "ZDATA");
                break;
        }

        /* NEW ^$EVENT */
        symtab (new_sym, "^$EVENT\201", "\201");
        
        /* populate ^$EVENT("GLVN") with the key */
        snprintf (k_buf, STRLEN - 1, "^$EVENT\202GLOBAL\201");
        symtab (set_sym, k_buf, d_buf);

        if ((action % 2) == 0) {

            snprintf (k_buf, STRLEN - 1, "^$EVENT\202OLD_VALUE\201");
            symtab (set_sym, k_buf, old_value);

            snprintf (k_buf, STRLEN - 1, "^$EVENT\202NEW_VALUE\201");
            symtab (set_sym, k_buf, new_value);

        }
        
        sprintf (ev_id, "%s:%s", an, r->name);
        evt_enqueue (ev_id, EVT_CLS_TRIGGER, 0);

        free (k_buf);
        free (d_buf);
    }

    free (old_value);
    free (new_value);
    free (r);
    free (tk_buf);
    free (old_ns);
    
    ierr = ierr_sav;
    
}
